import os
from datetime import timedelta, datetime
import re
import copy

iddict = {
    ('S5P_OFFL_L3CO____', '01_010400'): [('S5P_OFFL_L2__CO____', '01_010100'),
                                         ('S5P_OFFL_L2__CO____', '01_010200'),
                                         ('S5P_OFFL_L2__CO____', '01_010202'),
                                         ('S5P_OFFL_L2__CO____', '01_010[3-9][0-9][0-9]'),
                                         ('S5P_RPRO_L2__CO____', '01_010202'),
                                         ('S5P_RPRO_L2__CO____', '01_010[3-9][0-9][0-9]')],
    ('S5P_OFFL_L3CO____', '02_020200'): [('S5P_OFFL_L2__CO____', '02_02[0-9]{4}')],
    ('S5P_NRTI_L3CO____', '01_010400'): [('S5P_NRTI_L2__CO____', '01_010400')],
    ('S5P_NRTI_L3CO____', '02_020200'): [('S5P_NRTI_L2__CO____', '02_02[0-9]{4}')],

    ('S5P_OFFL_L3NO2___', '01_010400'): [('S5P_OFFL_L2__NO2___', '01_010200'),
                                         ('S5P_OFFL_L2__NO2___', '01_010202'),
                                         ('S5P_OFFL_L2__NO2___', '01_010[3-9][0-9][0-9]'),
                                         ('S5P_RPRO_L2__NO2___', '01_010202'),
                                         ('S5P_RPRO_L2__NO2___', '01_010[3-9][0-9][0-9]')],
    ('S5P_OFFL_L3NO2___', '02_020200'): [('S5P_OFFL_L2__NO2___', '02_02[0-9]{4}')],
    ('S5P_NRTI_L3NO2___', '01_010400'): [('S5P_NRTI_L2__NO2___', '[0-9]*')],
    ('S5P_NRTI_L3NO2___', '02_020200'): [('S5P_NRTI_L2__NO2___', '02_02[0-9]{4}')],

    ('S5P_OFFL_L3CH4___', '01_010202'): [('S5P_OFFL_L2__CH4___', '01_010202'),
                                         ('S5P_OFFL_L2__CH4___', '01_010[3-9][0-9][0-9]'),
                                         ('S5P_RPRO_L2__CH4___', '01_010202'),
                                         ('S5P_RPRO_L2__CH4___', '01_010[3-9][0-9][0-9]')],
    ('S5P_OFFL_L3HCHO__', '01_010105'): [('S5P_OFFL_L2__HCHO__', '01_01010[56789]'),
                                         ('S5P_RPRO_L2__HCHO__', '01_010105')]
}

tif_dict = {
    "CO": [{'nc_band_name': 'CO_column_volume_mixing_ratio_dry_air',  # used fore reading the nc-file
            'tif_band_name': 'CO',  # used in the tif file name
            'unit_conversion_factor' : 1,   #convert from mol to mol => no conversion for this band
            'scale': 10,
            "offset": 0,
            "dtype": 'int16',
            "nodata": 32767,
            "metadata": {
                "description": "Averaged carbon monoxide column volume mixing ratio for dry air [ppbv] based on Sentinel-5P/TROPOMI Level-2 %processing_mode% data.",
                "units": "[ppbv]"
            }
            },
           {'nc_band_name': 'weight',
            'tif_band_name': 'WEIGHT',  # used in the tif file name
            'unit_conversion_factor': 1,   #convert from mol to mol => no conversion for this band
            'scale': 1,
            "offset": 0,
            "dtype": 'int16',
            "nodata": -9999,
            "metadata": {
                "description": "weighting factors used for spatial binning based on Sentinel-5P/TROPOMI Level-2 %processing_mode% data.",
                "units": "-"
            }}
           ],
    "HCHO": [{'nc_band_name': 'tropospheric_HCHO_column_number_density',
              'tif_band_name': 'HCHO',  # used in the tif file name
              'unit_conversion_factor' : 1e-6,   #convert from mol to µmol
              'scale': 0.1,
              "offset": 0,
              "dtype": 'uint16',
              "nodata": 65535,
              "metadata": {
                  "description": "Tropospheric formaldehyde column number density [µmol m^-2] based on Sentinel-5P/TROPOMI Level-2 %processing_mode% data.",
                  "units": "[µmol m^-2]"
              }},
             {'nc_band_name': 'weight',
              'tif_band_name': 'WEIGHT',  # used in the tif file name
              'unit_conversion_factor' : 1,   #convert from mol to mol => no conversion for this band
              'scale': 1,
              "offset": 0,
              "dtype": 'int16',
              "nodata": 0,
              "metadata": {
                  "description": "weighting factors used for spatial binning based on Sentinel-5P/TROPOMI Level-2 %processing_mode% data.",
                  "units": "-"
              }}
             ],
    "NO2": [{'nc_band_name': 'tropospheric_NO2_column_number_density',
             'tif_band_name': 'NO2',  # used in the tif file name
             'unit_conversion_factor' : 1e-6,   #convert from mol to µmol
             'scale': 5,
             "offset": 0,
             "dtype": 'int16',
             "nodata": 32767,
             "metadata": {
                 "description": "Tropospheric vertical column of nitrogen dioxide [µmol m^-2] based on Sentinel-5P/TROPOMI Level-2 %processing_mode% data.",
                 "units": "[µmol m^-2]"
             }},
            {'nc_band_name': 'weight',
             'tif_band_name': 'WEIGHT',  # used in the tif file name
             'unit_conversion_factor' : 1,   #convert from mol to mol => no conversion for this band
             'scale': 1,
             "offset": 0,
             "dtype": 'int16',
             "nodata" : -9999,
             "metadata": {
                 "description": "weighting factors used for spatial binning based on Sentinel-5P/TROPOMI Level-2 %processing_mode% data.",
                 "units": "-"
             }}
            ]
}

class s5p_out_product(object):

    def __init__(self, content, logger, version, composit_type='TD' ):
        #### Definitions
        self.version = version
        self.latstep = 0.05
        self.lonstep = 0.05
        self.keep = 'keep(%s,latitude_bounds,longitude_bounds,datetime_start)'  # variables to keep when loading S5P file: only these are loaded into memory, %s is filled later on with the right target variable
        self.exclude_default = 'exclude(latitude_bounds_weight,longitude_bounds_weight,latitude,longitude)'
        self.oper_box = []
        self.oper_bin = 'bin_spatial(%d,-90,%.2f,%d,-180,%.2f)' % (
        int(180 / self.latstep) + 1, self.latstep, int(360 / self.lonstep + 1), self.lonstep)

        #setting parameters
        self.content = content
        self.processdate = content["day2process"]         #2020-01-01
        if composit_type == 'TD':
            self.processenddate = self.processdate
            tif_date = self.processdate.replace('-','')
        elif composit_type == 'TM':
            start_day = datetime.strptime(self.processdate,"%Y-%m-%d")
            self.processenddate = datetime.strftime((start_day.replace(day=1) + timedelta(days=32)).replace(day=1) - timedelta(days=1),'%Y-%m-%d')
            tif_date = self.processdate[:7].replace('-','') #202001
        elif composit_type == 'TY':
            start_day = datetime.strptime(self.processdate,"%Y-%m-%d")
            self.processenddate = datetime.strftime((start_day.replace(day=1) + timedelta(days=366)).replace(day=1) - timedelta(days=1),'%Y-%m-%d')
            tif_date = self.processdate[:4] #2020
        self.producttype = content["producttype"]           #CO
        self.processingmode = content["processingmode"]

        # copy the dict, not make another pointer to the same one
        self.tif_settings = copy.deepcopy(tif_dict[self.producttype])
        list(map(lambda x: x.update({'processingmode': self.processingmode }), self.tif_settings))


        #settings tifname
        tif_file = 'S5P_%s_%s_%s_<band>_V%s.tif' % (self.producttype, composit_type, tif_date, self.version)
        #update tif_settings with the tif_filename
        list(map(lambda x: x.update({'tif_file_name': tif_file}), self.tif_settings))
        #remove <band> from filenames and add the names to a list
        self.tif_file_lst = list(map(lambda x: tif_file.replace('<band>', x.get("tif_band_name")), self.tif_settings))

        self.processinglevel = 'L3'
        self.revisionnumber = content['revisionnumber']
        self.processorversion = self.revisionnumber  + '_' +content['processorversion']  #01_010400

        #self.processorversion = self.get_processorversion()

        self.composittype = composit_type                  #TD by default



        self.outfoldernc = content['outdir_nc']
        self.outfoldertif = content['outdir_tif']

        set_umask()
        #create output dir if not exist
        for folder in [self.outfoldernc, self.outfoldertif]:
            if not os.path.exists(folder):
                try:
                    os.makedirs(folder)
                    logger.info("-I- created folder %s " % folder)
                except:
                    logger.error("-E- unable to create dir %s" % folder)

        #create input objects, remove geometry for a beter overview
        self.inputs = []
        for jsoncontent in content['inputs']:
            index = content['inputs'].index(jsoncontent)
            self.content['inputs'][index]['geometry'] = ''
            logger.info("Adding product %s" % jsoncontent["title"])
            self.inputs.append(s5p_L2_product(jsoncontent) )
        self.nr_of_inputs = len(self.inputs)
        logger.info("-I- number of inputs %s " % self.nr_of_inputs)

        self.set_outfilename(logger)

        #setup/evaluate process parameters
        verified_inputs = self.filter_corrupt_inputs(logger)
        self.infile_str = ' '.join(verified_inputs)
        self.set_process_parameters(logger)

    # def get_processorversion(self):
    #     # S5P_OFFL_L3NO2___'
    #     key2search = 'S5P_%s_%s%s__' % (
    #         self.processingmode,
    #         self.processinglevel,
    #         self.producttype.ljust(4, '_')
    #     )
    #
    #     for tuple in iddict.keys():
    #         if tuple[0] == key2search:
    #             return tuple[1]
    #     return 'unknown'

    def set_outfilename(self, logger):
        # get datetime from file name component
        stimes=[];etimes=[]
        for input in self.inputs:
            s=input.beginposition
            stimes.append( datetime.strptime(s, '%Y-%m-%dT%H:%M:%SZ'))
        for input in self.inputs:
            s=input.endposition
            etimes.append(datetime.strptime(s, '%Y-%m-%dT%H:%M:%SZ'))
        out_stime = min(stimes)
        out_etime = max(etimes)
        out_proctime = datetime.now().strftime('%Y%m%dT%H%M%S')
        self.title = 'S5P_%s_L3_%s_%s_%s_V%s' % (
            self.processingmode,
            self.producttype,
            self.composittype,
            self.processdate.replace('-',''),
            self.version
        )
        self.outncfile = 'S5P_%s_L3%s%s___%s_%s_%s_%s.nc' %(
            self.processingmode,
            self.composittype,
            self.producttype,
            out_stime.strftime("%Y%m%dT%H%M%S"),
            out_etime.strftime("%Y%m%dT%H%M%S"),
            ''.join('{:05x}'.format(self.nr_of_inputs)),
            self.processorversion)  #S5P_OFFL_L2__CO_____20191207T002109_20191208T000205_{filecount}_01_010202.nc'
        self.outtiffile = self.outncfile.replace('.nc','.tif')
        logger.info("-I- outputfile : %s" % self.outtiffile)

    def set_process_parameters(self, logger):
        oper=[]
        exclude_default = 'exclude(latitude_bounds_weight,longitude_bounds_weight,latitude,longitude)'
        option = ''
        extra_harp_args = '--format netcdf --hdf5-compression 0 '
        keep = self.keep    #so we could copy/paste from terrascope3
        oper_bin = self.oper_bin    #so we could copy/paste from terrascope3

        logger.info("-I- producttype : %s" % self.producttype)

        if self.producttype == 'CO':
            self.nc_band_names=['CO_column_volume_mixing_ratio_dry_air','CO_column_volume_mixing_ratio_dry_air_count']
            if self.processingmode == 'OFFL':
                option='co=corrected'

            tvar = 'CO_column_number_density'
            oper.extend(['derive(CO_column_volume_mixing_ratio_dry_air {time} [ppbv])',
                         keep % ('CO_column_volume_mixing_ratio_dry_air'), self.oper_bin])
            # add filter, depends on processor
            if self.processorversion >= '010202':
                oper.insert(0, 'CO_column_number_density_validity >=70')  # mid level clouds
            else:
                oper = ['solar_zenith_angle <80', 'sensor_zenith_angle <65', 'cloud_height <5000',
                        'validity !& 4458240',
                        'scan_subindex >1'] + oper  # 4458240 for SRON ... 1073739775 more strict, better at wollongong

        elif self.producttype == 'NO2':
            self.nc_band_names=['tropospheric_NO2_column_number_density']
            tvar = 'tropospheric_NO2_column_number_density'
            oper.extend([keep % (','.join((tvar, 'scan_subindex', 'cloud_fraction'))), oper_bin])
            oper = ['tropospheric_NO2_column_number_density_validity>75'] + oper

        elif self.producttype == 'HCHO':
            self.nc_band_names=['tropospheric_HCHO_column_number_density']
            tvar = 'tropospheric_HCHO_column_number_density'
            oper.insert(0, '%s_validity >50' % tvar)
            oper.extend([keep % (tvar), oper_bin])

        else:
            self.nc_band_names=['']
            logger.error("-E- non existing producttype")
            logger.error("-E- type : %s" % self.producttype)
            raise ValueError('not existing producttype : %s ' % self.producttype)

        if self.composittype in ['TM','TY']:
            oper=[]
            oper.insert(0, exclude_default)

        if not isinstance(oper, str): oper = ';'.join(oper) #clean the list for empty strings, playing safe...
        self.harp_parameters = ' %s %s -%s "bin();exclude(latitude_bounds_weight,longitude_bounds_weight)" %s %s %s' % (
            ('-o "%s"' % option if option else ''),
            ('-a "%s"' % oper if oper else ''),
            'ar',
            extra_harp_args,
            self.infile_str,
            os.path.join(self.outfoldernc,self.outncfile)
            )


    def filter_corrupt_inputs(self, logger):
        self.check_processorversion()
        if self.composittype in ['TM','TY']:
            verified_inputs = self.check_if_L3_input_exists(logger)
        else:
            verified_inputs = self.check_if_L2_input_exists(logger)
        return verified_inputs

    def check_processorversion(self):
        #check if inputfiles have a version number that is high enoegh for the output
        for input in self.inputs:
            in_version = input.processorversion[-6:]
            if in_version > self.processorversion[-6:]:
                raise ValueError('inputfile version is to high for file %s : %s > %s' % (input.id, in_version, self.processorversion[-6:]))

    def check_if_L3_input_exists(self, logger):
        verified_inputs = []
        for input in self.inputs:
            in_path = input.fullpath.replace(".tif",'.nc')
            if not os.path.exists(in_path):
                logger.error("-E- input file (and sooner or later product) does not exist : %s" % in_path)
                raise ValueError('inputfile does not exist on the path : %s' % in_path)
            else:
                verified_inputs.append(in_path)
        return verified_inputs



    def check_if_L2_input_exists(self, logger):
        verified_inputs=[]
        for input in self.inputs:
            in_path = input.fullpath
            if not os.path.exists(in_path):
                #check on next/previous day
                pattern = '/([0-9]{4}/[0-9]{2}/[0-9]{2})/'
                m = re.search(pattern, in_path)
                date_folder_structure = m.group(1)
                begindate = datetime.strptime(date_folder_structure, '%Y/%m/%d')
                plusdate = begindate + timedelta(days=1)
                mindate = begindate - timedelta(days=1)
                pluspath = in_path.replace(date_folder_structure, plusdate.strftime('%Y/%m/%d'))
                minpath = in_path.replace(date_folder_structure, mindate.strftime('%Y/%m/%d'))
                if os.path.exists(pluspath):
                    verified_inputs.append(pluspath)
                    logger.info("Using product from next day directory %s" % pluspath)
                elif os.path.exists(minpath):
                    verified_inputs.append(minpath)
                    logger.info("Using product from previous day directory %s" % minpath)
                elif os.path.exists(in_path.replace(".nc",'.zip')):
                    verified_inputs.append(in_path.replace(".nc",'.zip'))
                    logger.info("hack. switch extension from nc to zip : %s" % in_path)
                else:
                    logger.error("-E- input file (and sooner or later product) does not exist : %s" % in_path)
                    raise ValueError('inputfile does not exist on the path : %s' % in_path)
            else:
                verified_inputs.append(in_path)
        return verified_inputs


    def update_metadata_json(self):
        content = self.content
        content['composittype'] = self.composittype
        content['processingmode'] = self.processingmode
        content['title'] = self.title
        content['outname'] = self.outncfile
        content['nr_of_inputs'] = self.nr_of_inputs
        content['processorversion'] = self.processorversion
        content['version']=self.version
        content['revisionnumber'] = self.revisionnumber
        content['processinglevel'] = self.processinglevel
        content['beginposition'] = self.processdate + 'T00:00:00Z'
        content['endposition']   = self.processenddate + 'T23:59:59Z'
        content['collection'] = 'S5P_%s_%s_%s_%s_%s' %(
            self.processingmode,
            self.processinglevel,
            self.producttype,
            self.composittype,
            self.version
        ) #L3_NO2_TD_V100
        content['producttype'] = '%s_%s' % (self.producttype , self.composittype) #NO2_TD
        content['output_tif_list'] = self.tif_file_lst

        self.content = content

    def set_metadata(self,key,value):
        self.content[key]= value

    def update_filesize_json(self):
        p = os.path.join(self.outfoldertif, self.outtiffile)
        if os.path.exists(p):
            self.content['filesize'] = os.path.getsize(p)
        else: self.content['filesize'] = 'Nan'


class s5p_L2_product(object):

    def __init__(self, content):

        self.id = content["title"]

        properties = content['properties']
        self.fullpath = properties["fullpath"]
        self.beginposition = properties['beginposition']
        self.endposition = properties['endposition']
        self.processingmode = properties['processingmode']
        self.processorversion  = properties['processorversion']
        self.processinglevel = properties['processinglevel']
        self.producttype = properties["producttype"]
        if "revisionnumber" in properties:
            self.revisionnumber = properties["revisionnumber"]
        else:
            self.revisionnumber = self.processorversion.split('_')[0]

def set_umask():
    """Set umask for right permissions"""
    try:
        os.umask(0o000)
    except:
        pass